"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("../utils");
const QueryBuilderHelper_1 = require("./QueryBuilderHelper");
const entity_1 = require("../entity");
const enums_1 = require("./enums");
/**
 * SQL query builder
 */
class QueryBuilder {
    constructor(entityName, metadata, driver) {
        this.entityName = entityName;
        this.metadata = metadata;
        this.driver = driver;
        this.alias = `e0`;
        this.aliasCounter = 1;
        this.flags = [];
        this.finalized = false;
        this._populate = [];
        this._populateMap = {};
        this._leftJoins = {};
        this._cond = {};
        this.connection = this.driver.getConnection();
        this.platform = this.driver.getPlatform();
        this.helper = new QueryBuilderHelper_1.QueryBuilderHelper(this.entityName, this.alias, this.metadata, this.platform);
    }
    select(fields) {
        this._fields = Array.isArray(fields) ? fields : [fields];
        return this.init(enums_1.QueryType.SELECT);
    }
    insert(data) {
        return this.init(enums_1.QueryType.INSERT, data);
    }
    update(data) {
        return this.init(enums_1.QueryType.UPDATE, data);
    }
    delete(cond = {}) {
        return this.init(enums_1.QueryType.DELETE, undefined, cond);
    }
    truncate() {
        return this.init(enums_1.QueryType.TRUNCATE);
    }
    count(field, distinct = false) {
        this.select(field || this.metadata[this.entityName].primaryKey);
        this.flags.push(enums_1.QueryFlag.COUNT);
        if (distinct) {
            this.flags.push(enums_1.QueryFlag.DISTINCT);
        }
        return this;
    }
    where(cond, operator) {
        if (!operator) {
            this._cond = this.processWhere(cond);
        }
        else if (Array.isArray(this._cond[operator])) {
            this._cond[operator].push(this.processWhere(cond));
        }
        else {
            this._cond = { [operator]: [this._cond, this.processWhere(cond)] };
        }
        return this;
    }
    andWhere(cond) {
        return this.where(cond, '$and');
    }
    orWhere(cond) {
        return this.where(cond, '$or');
    }
    orderBy(orderBy) {
        orderBy = Object.assign({}, orderBy); // copy first
        Object.keys(orderBy).forEach(field => {
            if (!this.metadata[this.entityName] || !this.metadata[this.entityName].properties[field]) {
                return;
            }
            const prop = this.metadata[this.entityName].properties[field];
            utils_1.Utils.renameKey(orderBy, field, prop.fieldName);
        });
        this._orderBy = orderBy;
        return this;
    }
    populate(populate) {
        this._populate = populate;
        return this;
    }
    limit(limit, offset = 0) {
        this._limit = limit;
        if (offset) {
            this.offset(offset);
        }
        return this;
    }
    offset(offset) {
        this._offset = offset;
        return this;
    }
    getQuery() {
        this.finalize();
        let sql = this.getQueryBase();
        if (this._cond && Object.keys(this._cond).length > 0) {
            sql += ' WHERE ' + this.helper.getQueryCondition(this.type, this._cond).join(' AND ');
        }
        if (this._orderBy && Object.keys(this._orderBy).length > 0) {
            sql += ' ORDER BY ' + this.helper.getQueryOrder(this.type, this._orderBy, this._populateMap).join(', ');
        }
        sql += this.helper.getQueryPagination(this._limit, this._offset);
        if (this.type === enums_1.QueryType.TRUNCATE && this.platform.usesCascadeStatement()) {
            sql += ' CASCADE';
        }
        return this.helper.finalize(this.type, sql, this.metadata[this.entityName]);
    }
    getParams() {
        this.finalize();
        let ret = [];
        if (this.type === enums_1.QueryType.INSERT && this._data) {
            ret = Object.values(this._data);
        }
        else if (this.type === enums_1.QueryType.UPDATE) {
            ret = Object.values(this._data);
        }
        if (this._cond) {
            ret = ret.concat(this.helper.getWhereParams(this._cond));
        }
        if (this._limit) {
            ret.push(this._limit);
        }
        if (this._offset) {
            ret.push(this._offset);
        }
        return ret;
    }
    async execute(method = 'all', mapResults = true) {
        const res = await this.connection.execute(this.getQuery(), this.getParams(), method);
        if (!mapResults) {
            return res;
        }
        if (method === 'all' && Array.isArray(res)) {
            return res.map((r) => this.driver.mapResult(r, this.metadata[this.entityName]));
        }
        return this.driver.mapResult(res, this.metadata[this.entityName]);
    }
    prepareFields(fields, glue = ', ') {
        const ret = [];
        fields.forEach(f => {
            if (this._leftJoins[f]) {
                ret.push(...this.helper.mapJoinColumns(this.type, this._leftJoins[f]));
                return;
            }
            ret.push(this.helper.mapper(this.type, f));
        });
        Object.keys(this._populateMap).forEach(f => {
            if (!fields.includes(f)) {
                ret.push(...this.helper.mapJoinColumns(this.type, this._leftJoins[f]));
            }
            if (this._leftJoins[f].prop.reference !== entity_1.ReferenceType.ONE_TO_ONE) {
                utils_1.Utils.renameKey(this._cond, this._leftJoins[f].inverseJoinColumn, `${this._leftJoins[f].alias}.${this._leftJoins[f].inverseJoinColumn}`);
            }
        });
        if (this.flags.includes(enums_1.QueryFlag.COUNT)) {
            if (this.flags.includes(enums_1.QueryFlag.DISTINCT)) {
                return `COUNT(DISTINCT ${ret[0]}) AS ${this.helper.wrap('count')}`;
            }
            return `COUNT(${ret[0]}) AS ${this.helper.wrap('count')}`;
        }
        return ret.join(glue);
    }
    processWhere(cond) {
        cond = Object.assign({}, cond); // copy first
        Object.keys(cond).forEach(field => {
            this.helper.replaceEmptyInConditions(cond, field);
            if (!this.metadata[this.entityName] || !this.metadata[this.entityName].properties[field]) {
                return;
            }
            const prop = this.metadata[this.entityName].properties[field];
            if (prop.reference === entity_1.ReferenceType.ONE_TO_MANY || prop.reference === entity_1.ReferenceType.MANY_TO_MANY) {
                this.processCollection(prop, cond);
            }
            else if (prop.reference === entity_1.ReferenceType.ONE_TO_ONE) {
                this.processOneToOne(prop, cond);
            }
            else {
                utils_1.Utils.renameKey(cond, field, prop.fieldName);
            }
        });
        return cond;
    }
    processOneToOne(prop, cond) {
        if (prop.owner) {
            return utils_1.Utils.renameKey(cond, prop.name, prop.fieldName);
        }
        this._fields.push(prop.name);
        const prop2 = this.metadata[prop.type].properties[prop.mappedBy];
        const alias2 = this.joinOneToReference(prop, prop2);
        utils_1.Utils.renameKey(cond, prop.name, `${alias2}.${prop2.referenceColumnName}`);
    }
    processCollection(prop, cond) {
        if (prop.reference === entity_1.ReferenceType.MANY_TO_MANY) {
            const alias1 = `e${this.aliasCounter++}`;
            const join = {
                alias: alias1,
                joinColumn: prop.joinColumn,
                inverseJoinColumn: prop.inverseJoinColumn,
                primaryKey: prop.referenceColumnName,
            };
            if (prop.owner) {
                this._leftJoins[prop.name] = Object.assign(join, { table: prop.pivotTable });
            }
            else {
                const prop2 = this.metadata[prop.type].properties[prop.mappedBy];
                this._leftJoins[prop.name] = Object.assign(join, { table: prop2.pivotTable });
            }
            this._fields.push(prop.name);
            utils_1.Utils.renameKey(cond, prop.name, `${alias1}.${prop.inverseJoinColumn}`);
        }
        else if (prop.reference === entity_1.ReferenceType.ONE_TO_MANY) {
            const prop2 = this.metadata[prop.type].properties[prop.mappedBy];
            const alias2 = this.joinOneToReference(prop, prop2);
            utils_1.Utils.renameKey(cond, prop.name, `${alias2}.${prop.referenceColumnName}`);
        }
    }
    joinOneToReference(prop, prop2) {
        const alias2 = `e${this.aliasCounter++}`;
        this._leftJoins[prop.name] = {
            table: this.helper.getTableName(prop.type),
            alias: alias2,
            joinColumn: prop2.fieldName,
            inverseJoinColumn: prop2.referenceColumnName,
            primaryKey: prop.referenceColumnName,
            prop,
        };
        return alias2;
    }
    init(type, data, cond) {
        this.type = type;
        if (data) {
            this._data = this.helper.processData(data);
        }
        if (cond) {
            this._cond = this.processWhere(cond);
        }
        return this;
    }
    getQueryBase() {
        let sql = this.type + ' ';
        switch (this.type) {
            case enums_1.QueryType.SELECT:
                sql += this.prepareFields(this._fields);
                sql += ` FROM ${this.helper.getTableName(this.entityName, true)} AS ${this.helper.wrap(this.alias)}`;
                sql += this.helper.processJoins(this._leftJoins);
                break;
            case enums_1.QueryType.INSERT:
                sql += `INTO ${this.helper.getTableName(this.entityName, true)}`;
                sql += ' (' + Object.keys(this._data).map(k => this.helper.wrap(k)).join(', ') + ')';
                sql += ' VALUES (' + Object.keys(this._data).map(() => '?').join(', ') + ')';
                break;
            case enums_1.QueryType.UPDATE:
                sql += this.helper.getTableName(this.entityName, true);
                sql += ' SET ' + Object.keys(this._data).map(k => this.helper.wrap(k) + ' = ?').join(', ');
                break;
            case enums_1.QueryType.DELETE:
                sql += 'FROM ' + this.helper.getTableName(this.entityName, true);
                break;
            case enums_1.QueryType.TRUNCATE:
                sql += 'TABLE ' + this.helper.getTableName(this.entityName, true);
                break;
        }
        return sql;
    }
    finalize() {
        if (this.finalized) {
            return;
        }
        this._populate.forEach(field => {
            if (this._leftJoins[field]) {
                return this._populateMap[field] = this._leftJoins[field].alias;
            }
            if (this.metadata[field]) { // pivot table entity
                const prop = this.metadata[field].properties[this.entityName];
                this._leftJoins[field] = {
                    table: this.metadata[field].collection,
                    alias: `e${this.aliasCounter++}`,
                    joinColumn: prop.joinColumn,
                    inverseJoinColumn: prop.inverseJoinColumn,
                    primaryKey: prop.referenceColumnName,
                    prop,
                };
                this._populateMap[field] = this._leftJoins[field].alias;
            }
            else if (this.helper.isOneToOneInverse(field)) {
                const prop = this.metadata[this.entityName].properties[field];
                const prop2 = this.metadata[prop.type].properties[prop.mappedBy];
                this.joinOneToReference(prop, prop2);
                this._populateMap[field] = this._leftJoins[field].alias;
            }
        });
        this.finalized = true;
    }
}
exports.QueryBuilder = QueryBuilder;
